#!/usr/bin/env ruby
require 'timeout'
require 'tempfile'


# functions

def read_commands(file)
 cmds = Array.new
 i=0
 File.new(file,"r").each do |cmd|
  cmds[i] =cmd[0,cmd.index("\n")]+" "
  i=i+1
 end
 return cmds
end

def read_output(outputfile)
  result = Array.new
  file = File.new(outputfile, "r")
  line = file.gets
  if (line=~/YES/ ) then status = 0
  elsif (line=~/MAYBE/ ) then status = 1
  else status = 2
  end
#  puts "status is #{String(status)}"
  line = file.gets
  line=~/Time:\s*(\d*\.\d*).*/
#  puts "line is '#{line}'"
  time = if (status<=1) then $1 else -1 end
  file.close
  return [status,time]
end

def html_cell(dir,file,i)
 data = read_output(dir+file)
 puts "Result: #{String(data[0])}, #{String(data[1])}"
 if (data[0] == 0) then # success
  css_class="yes"
  time = String(data[1])
  $successes[i] += 1
  $times[i] += data[1].to_f
 elsif (data[0] == 1) then # timeout
  css_class="timeout"
  time = String(data[1])
  $times[i] += data[1].to_f
 else # failure
  css_class="error"
  time = "&perp;"
 end
 style="font-family: Arial, Helvetica, sans-serif; font-size: 8pt"
 cell="<td class=\"#{css_class}\">"
 cell+="<a href=\"#{file}\">#{time}</a></td>\n"
end

#main
inputdir = ARGV[0]
commands_file = ARGV[1]
outputdir = ARGV[2]
$successes = Array.new
$times = Array.new
timeout=60
filecount = 0

if not outputdir =~/\/$/ then outputdir=outputdir+"/" end

# read commands
cmds = read_commands(commands_file)
#create output directories
i=0
cmds.each do
 if !(File.exist?(outputdir + String(i))) then
  Dir.mkdir(outputdir + String(i))
 end
 $successes[i]=0
 $times[i]=0
 i=i+1
end
# prepare html
html = "<html>\n"
html+="<link rel=\"stylesheet\" type=\"text/css\" href=\"http://colo1-c703.uibk.ac.at/workspace/style_experiments.css\">\n"
html+="<link rel=\"stylesheet\" type=\"text/css\" href=\"http://colo1-c703.uibk.ac.at/workspace/style-wide.css\">\n"
html += "<table cellspacing=\"1\" cellpadding=\"1\" border=\"0\" class=\"experiments\" style=\"font-family: Arial, Helvetica, sans-serif; font-size: 8pt; text-align:center\">\n"
html+="<tr><th>&nbsp;</th>"
for j in 0..(i-1)
      html += "<th width=\"30\">#{String(j)}</th>"
end
html +="</tr>"
# run commands
Dir.new(inputdir).sort.each do |filename|
  if not File.directory?(filename) then
    if filename =~ /^(.*)\.([^\.]*?)$/o then
     basename=$1
     suffix=$2
     filecount += 1
    else
      puts "Unknown file type "+filename
      continue
    end
    # file split into basename, suffix
    trs = "#{inputdir}/#{basename}.#{suffix}"
    puts "Processing "+trs
    # loop over commands
    i=0
    pids = Array.new
    cmds.each do |cmd|
     cmdc = "./sandbox #{timeout} "+cmd + "#{trs} > #{outputdir}#{String(i)}/#{basename}.txt"
     puts cmdc
     pids[i] = Process.fork do                                                 
       system(cmdc)
     end
     i=i+1
    end # loop over cmds
    #loop over wait
    pids.each do |pid|
     Process.waitpid(pid)
    end

    # construct html
    html += "<tr><td class= \"copyright\"><a href=\"../../#{trs}\">#{basename}</a></td>\n"
    for j in 0..(i-1)
      html += html_cell(outputdir, "#{String(j)}/#{basename}.txt",j)
    end
    html +="</tr>\n"
  end # no directory
end # file loop
# result summary
i=0
# number of successes
html += "<tr><td class= \"copyright\">successes</td>"
for i in 0..(cmds.length - 1)
 html += "<td class= \"copyright\">#{String($successes[i])}</td>\n"
end
html += "</tr>"
# total time
html += "<tr><td class= \"copyright\">total time</td>"
for i in 0..(cmds.length - 1)
 t = $times[i]#+(filecount - ($successes[i]))*timeout
 html += "<td class= \"copyright\">#{String(t)}</td>\n"
end
html += "</tr>"
# average time
#html += "<tr><td class= \"copyright\">&empty; time</td>"
#for i in 0..(cmds.length - 1)
# s = ($successes[i] == 0) ? 0.001 : $successes[i]
# html += "<td class= \"copyright\">#{String($times[i]/$successes[i])}</td>\n"
#end
html += "</tr>"
#finalize html
html += "</table><p><p>\n"
# add legend
html += "<table cellspacing=\"1\" cellpadding=\"0\" border=\"0\" class=\"experiments\" style=\"font-family: Arial, Helvetica, sans-serif; font-size: 8pt\">\n"
html += "<tr><td colspan=\"2\" style=\"color: #6060D4;\">Legend</td></tr>\n"
i=0
cmds.each do |cmd|
 html += "<tr><td style=\"color: #6060D4;\">#{String(i)}</td><td><code>#{cmd}</code></td></tr>\n"
 i=i+1
end
html += "</table>"
html+="</html>"
File.open("#{outputdir}results.html", 'w') {|f| f.write(html) }
puts html
